Package: mbkmeans
Type: Package
Title: Mini-batch K-means Clustering for Single-Cell RNA-seq
Version: 1.28.0
Authors@R: 
    c(person("Yuwei", "Ni", role = c("aut", "cph"),
             email = "yuweini45@gmail.com"),
      person("Davide", "Risso", role = c("aut", "cre", "cph"),
             email = "risso.davide@gmail.com"),
      person("Stephanie", "Hicks", role = c("aut", "cph"),
             email = "shicks19@jhu.edu"),
      person("Elizabeth", "Purdom", role = c("aut", "cph"),
             email = "epurdom@stat.berkeley.edu"))
Description: 
    Implements the mini-batch k-means algorithm for large
    datasets, including support for on-disk data representation.
Depends: R (>= 3.6)
Imports: methods, DelayedArray, Rcpp, S4Vectors, SingleCellExperiment,
        SummarizedExperiment, ClusterR, benchmarkme, Matrix,
        BiocParallel
Suggests: beachmat, HDF5Array, Rhdf5lib, BiocStyle, TENxPBMCData,
        scater, DelayedMatrixStats, bluster, knitr, testthat, rmarkdown
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.1.1
LinkingTo: Rcpp, RcppArmadillo (>= 0.7.2), Rhdf5lib, beachmat, ClusterR
SystemRequirements: C++11
VignetteBuilder: knitr
biocViews: Clustering, GeneExpression, RNASeq, Software,
        Transcriptomics, Sequencing, SingleCell
BugReports: https://github.com/drisso/mbkmeans/issues
git_url: https://git.bioconductor.org/packages/mbkmeans
git_branch: RELEASE_3_23
git_last_commit: a415e0d
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: yes
Packaged: 2026-04-28 23:04:51 UTC; biocbuild
Author: Yuwei Ni [aut, cph],
  Davide Risso [aut, cre, cph],
  Stephanie Hicks [aut, cph],
  Elizabeth Purdom [aut, cph]
Maintainer: Davide Risso <risso.davide@gmail.com>

# Generated by roxygen2: do not edit by hand

export(MbkmeansParam)
export(blocksize)
export(compute_wcss)
export(mini_batch)
export(predict_mini_batch)
export(predict_mini_batch_r)
exportMethods(mbkmeans)
import(ClusterR)
importClassesFrom(DelayedArray,DelayedMatrix)
importClassesFrom(SingleCellExperiment,LinearEmbeddingMatrix)
importClassesFrom(SingleCellExperiment,SingleCellExperiment)
importClassesFrom(SummarizedExperiment,SummarizedExperiment)
importFrom(BiocParallel,SerialParam)
importFrom(DelayedArray,blockApply)
importFrom(DelayedArray,colAutoGrid)
importFrom(DelayedArray,colSums)
importFrom(Matrix,t)
importFrom(Rcpp,sourceCpp)
importFrom(SingleCellExperiment,reducedDim)
importFrom(SingleCellExperiment,reducedDimNames)
importFrom(SingleCellExperiment,sampleFactors)
importFrom(SummarizedExperiment,assay)
importFrom(SummarizedExperiment,assays)
importFrom(benchmarkme,get_ram)
importFrom(methods,is)
useDynLib(mbkmeans, .registration = TRUE)

#' Mini-batch k-means clustering
#'
#' Run the mini-batch k-means \code{\link{mbkmeans}} function with the specified
#' number of centers within \code{clusterRows} from the
#' \code{bluster} Bioconductor package.
#'
#' This function is deprecated. Please use the \code{MbkmeansParam} function in
#' the \code{bluster} Bioconductor package.
#' 
#' @param centers An integer scalar specifying the number of centers.
#' Alternatively, a function that takes the number of observations and returns the number of centers. 
#' Note, the \code{\link{mbkmeans}} function uses the argument \code{clusters} argument to represent this argument. 
#' However, we use \code{centers} to match 
#' @param ... Further arguments to pass to \code{\link{mbkmeans}}.
#' 
#' @name MbkmeansParam
#' @docType class
#' @aliases 
#' show,MbkmeansParam-method
NULL

#' @export
#' @rdname MbkmeansParam
MbkmeansParam <- function(centers, ...) {
  .Deprecated("bluster::MbkmeansParam")
  bluster::MbkmeansParam(centers, ...)
}

#' @importFrom Rcpp sourceCpp
#' @import ClusterR
#' @useDynLib mbkmeans, .registration = TRUE
#'
NULL


#' @title Mini-Batch k-means for large single cell sequencing data
#'
#' @description This is an implementation of the mini-batch k-means algorithm of
#'   Sculley (2010) for large single cell sequencing data with the
#'   dimensionality reduction results as input in the reducedDim() slot.
#'
#' @details The implementation is largely based on the
#'   \code{\link[ClusterR]{MiniBatchKmeans}} function of the \code{ClusterR}
#'   package. The contribution of this package is to provide support for on-disk
#'   data representations such as HDF5, through the use of \code{DelayedMatrix}
#'   and \code{HDF5Matrix} objects, as well as for sparse data representation
#'   through the classes of the \code{Matrix} package. We also provide
#'   high-level methods for objects of class \code{SummarizedExperiment},
#'   \code{SingleCellExperiment}, and \code{LinearEmbeddingMatrix}.
#'
#' @param x The object on which to run mini-batch k-means. It can be a
#'   matrix-like object (e.g., matrix, Matrix, DelayedMatrix, HDF5Matrix) with
#'   genes in the rows and samples in the columns. Specialized methods are
#'   defined for SummarizedExperiment and SingleCellExperiment.
#' @param ... Arguments to pass to the matrix method.
#' @return A list with the following attributes: centroids, WCSS_per_cluster,
#'   best_initialization, iters_per_initialization.
#' @rdname mbkmeans
#' @importFrom methods is
#' @importFrom SummarizedExperiment assay
#' @export
#' @importClassesFrom SummarizedExperiment SummarizedExperiment
#' @references Sculley. Web-Scale K-Means Clustering. WWW 2010, April 26–30,
#'   2010, Raleigh, North Carolina, USA. ACM 978-1-60558-799-8/10/04.
#' @author Lampros Mouselimis and Yuwei Ni
#' @examples
#' library(SummarizedExperiment)
#' se <- SummarizedExperiment(matrix(rnorm(100), ncol=10))
#' mbkmeans(se, clusters = 2)
setMethod(
    f = "mbkmeans",
    signature = signature(x = "SummarizedExperiment"),
    definition = function(x, whichAssay = 1, ...){
        mbkmeans(assay(x, whichAssay), ...)
})

#' @rdname mbkmeans
#' @export
#' @importClassesFrom SingleCellExperiment SingleCellExperiment
#' @importFrom SummarizedExperiment assays
#' @importFrom SingleCellExperiment reducedDim reducedDimNames
#' @importFrom Matrix t
#' @param reduceMethod Name of dimensionality reduction results to use as input
#'   to mini-batch k-means. Set to NA to use the full matrix.
#' @param whichAssay The assay to use as input to mini-batch k-means. If x is a
#'   SingleCellExperiment, this is ignored unless \code{reduceMethod = NA}.
#' @examples
#' library(SingleCellExperiment)
#' sce <- SingleCellExperiment(matrix(rnorm(100), ncol=10))
#' mbkmeans(sce, clusters = 2, reduceMethod = NA)
setMethod(
    f = "mbkmeans",
    signature = signature(x = "SingleCellExperiment"),
    definition = function(x, reduceMethod = "PCA", whichAssay = 1, ...)
    {

        if(is.na(reduceMethod)){
            if(NCOL(x)>10000)
                message("Note that you are running kmeans with more than",
                        "10,000 cells using all of the dimensions.", "\n",
                        "You might consider running a",
                        "dimensionality reduction step first.")
            fit <- mbkmeans(assay(x, whichAssay), ...)
        }
        else{
            if(is.null(reducedDimNames(x))){
                stop("There are no dimensionality reduction results",
                "stored in this object. Use reduceDims() to store",
                "dimensionality reduction results.")
            }
            if(!(reduceMethod %in% reducedDimNames(x))){
                stop("The argument reduceMethod does not match one",
                "of the reducedDimNames() in this object. Use",
                "reducedDimNames() to see what names are in this object.")
            }
            fit <- mbkmeans(t(reducedDim(x, reduceMethod)), ...)

        }

        return(fit)
    })

#' @rdname mbkmeans
#' @export
#' @importClassesFrom SingleCellExperiment LinearEmbeddingMatrix
#' @importFrom SingleCellExperiment sampleFactors
setMethod(
    f = "mbkmeans",
    signature = signature(x = "LinearEmbeddingMatrix"),
    definition = function(x, ...)
    {
        mbkmeans(t(sampleFactors(x)), ...)
})

#'@rdname mbkmeans
#'@export
#'@importClassesFrom DelayedArray DelayedMatrix
#'@param clusters the number of clusters
#'@param batch_size the size of the mini batches. By default, it equals the
#'  minimum between the number of observations and 500.
#'@param num_init number of times the algorithm will be run with different
#'  centroid seeds
#'@param max_iters the maximum number of clustering iterations
#'@param init_fraction proportion of data to use for the initialization
#'  centroids (applies if initializer is \emph{kmeans++} ). Should be a float
#'  number between 0.0 and 1.0. By default, it uses the relative batch size.
#'@param initializer the method of initialization. One of \emph{kmeans++} and
#'  \emph{random}. See details for more information
#'@param early_stop_iter continue that many iterations after calculation of the
#'  best within-cluster-sum-of-squared-error
#'@param compute_labels logcical indicating whether to compute the final cluster
#'  labels.
#'@param calc_wcss logical indicating whether the per-cluster WCSS
#'  is computed. Ignored if `compute_labels = FALSE`.
#'@param verbose either TRUE or FALSE, indicating whether progress is printed
#'  during clustering
#'@param CENTROIDS a matrix of initial cluster centroids. The rows of the
#'  CENTROIDS matrix should be equal to the number of clusters and the columns
#'  should be equal to the columns of the data
#'@param tol a float number. If, in case of an iteration (iteration > 1 and
#'  iteration < max_iters) 'tol' is greater than the squared norm of the
#'  centroids, then kmeans has converged
#'@param BPPARAM See the `BiocParallel` package. Only the label assignment is
#'  done in parallel.
#'@param ... passed to `blockApply`.
#'@return a list with the following attributes: centroids, WCSS_per_cluster,
#'  best_initialization, iters_per_initialization
#'@details This function performs k-means clustering using mini batches.
#'
#'\strong{kmeans++}: kmeans++ initialization. Reference :
#'http://theory.stanford.edu/~sergei/papers/kMeansPP-soda.pdf AND
#'http://stackoverflow.com/questions/5466323/how-exactly-does-k-means-work
#'
#'\strong{random}: random selection of data rows as initial centroids
#'
#'@references https://github.com/mlampros/ClusterR
#'
#'@examples
#'x<-matrix(rnorm(100), ncol=10)
#'mbkmeans(x,clusters = 3)
#'
setMethod(
    f = "mbkmeans",
    signature = signature(x ="ANY"),
    definition = function(x, clusters,
                        batch_size = min(500, NCOL(x)),
                        max_iters =100, num_init = 1,
                        init_fraction = batch_size / NCOL(x),
                        initializer = "kmeans++", compute_labels = TRUE,
                        calc_wcss = FALSE, early_stop_iter = 10,
                        verbose = FALSE,
                        CENTROIDS = NULL, tol = 1e-4,
                        BPPARAM = BiocParallel::SerialParam(),
                        ...)
    {

        if(!is(x, "matrix") && !is(x, "Matrix") && !is(x, "HDF5Matrix") &&
           !is(x, "DelayedMatrix")) {
            stop("x is of type ", class(x), ", currently not supported")

        } else {
            fit <- mini_batch(data = t(x), clusters = clusters,
                              batch_size = batch_size, max_iters = max_iters,
                              num_init = num_init,
                              init_fraction = init_fraction,
                              initializer = initializer,
                              compute_labels = FALSE,
                              calc_wcss = FALSE,
                              early_stop_iter = early_stop_iter,
                              verbose = verbose,
                              CENTROIDS = CENTROIDS, tol = tol)
            if(compute_labels) {
                fit$Clusters <- predict_mini_batch_r(x, fit$centroids,
                                                     BPPARAM = BPPARAM, ...)

                if(calc_wcss) {
                    fit$WCSS_per_cluster <- compute_wcss(fit$Clusters,
                                                         fit$centroids,
                                                         t(x))
                }
            }
        }

        return(fit)
    })

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' Predict_mini_batch
#'
#' Prediction function for mini-batch k-means applied to matrix-like objects.
#'
#'
#'@param data matrix-like objectcontaining numeric or
#'  integer data (obseravtions in rows, variables in columns).
#'@param CENTROIDS a matrix of initial cluster centroids. The rows of the
#'  CENTROIDS matrix should be equal to the number of clusters and the columns
#'  should equal the columns of the data.
#'@return it returns a vector with the clusters.
#'@details
#'
#'This function takes the data and the output centroids and returns the
#'clusters.
#'
#'This implementation relies very heavily on the
#'\code{\link[ClusterR]{MiniBatchKmeans}} implementation. We provide the
#'ability to work with other matrix-like objects other than base matrices (e.g,
#'DelayedMatrix and HDF5Matrix) through the \code{beachmat}
#'library.
#'
#'@author Yuwei Ni
#'
#'@examples
#'data(iris)
#'km = mini_batch(as.matrix(iris[,1:4]), clusters = 3,
#'                batch_size = 10, max_iters = 10)
#'clusters = predict_mini_batch(as.matrix(iris[,1:4]),
#'                              CENTROIDS = km$centroids)
#' @export
predict_mini_batch <- function(data, CENTROIDS) {
    .Call(`_mbkmeans_predict_mini_batch`, data, CENTROIDS)
}

#' Compute Whithin-Cluster Sum of Squares
#'
#' Given a vector of cluster labels, a matrix of centroids, and a dataset, it
#' computes the WCSS.
#'
#'@param clusters numeric vector with the cluster assignments.
#'@param cent numeric matrix with the centroids (clusters in rows, variables
#'  in columns).
#'@param data matrix-like object containing the data (numeric or integer).
#'
#'@return A numeric vector with the value of WCSS per cluster.
#'
#'@examples
#'data = matrix(1:30,nrow = 10)
#'cl <- mini_batch(data, 2, 10, 10)
#'compute_wcss(cl$Clusters, cl$centroids, data)
#'
#' @export
compute_wcss <- function(clusters, cent, data) {
    .Call(`_mbkmeans_compute_wcss`, clusters, cent, data)
}

#' Mini_batch
#'
#' Mini-batch-k-means for matrix-like objects
#'
#'@param data numeric or integer matrix-like object.
#'@param clusters the number of clusters.
#'@param batch_size the size of the mini batches.
#'@param num_init number of times the algorithm will be run with different
#'  centroid seeds.
#'@param max_iters the maximum number of clustering iterations.
#'@param init_fraction percentage of data to use for the initialization
#'  centroids (applies if initializer is \emph{kmeans++} ). Should be a float
#'  number between 0.0 and 1.0.
#'@param initializer the method of initialization. One of \emph{kmeans++} and
#'  \emph{random}. See details for more information.
#'@param compute_labels logical indicating whether to compute the final cluster
#'  labels.
#'@param calc_wcss logical indicating whether the within-cluster sum of squares
#'  should be computed and returned (ignored if `compute_labels = FALSE`).
#'@param early_stop_iter continue that many iterations after calculation of the
#'  best within-cluster-sum-of-squared-error.
#'@param verbose logical indicating whether progress is printed on screen.
#'@param CENTROIDS an optional matrix of initial cluster centroids. The rows of
#'  the CENTROIDS matrix should be equal to the number of clusters and the
#'  columns should be equal to the columns of the data.
#'@param tol convergence tolerance.
#'@return
#'a list with the following attributes:
#'
#'centroids: the final centroids;
#'
#'WCSS_per_cluster (optional): the final per-cluster WCSS.
#'
#'best_initialization: which initialization value led to the best WCSS
#'solution;
#'
#'iters_per_initialization: number of iterations per each initialization;
#'
#'Clusters (optional): the final cluster labels.
#'
#'@details This function performs k-means clustering using mini batches. It was
#'inspired by the implementation in https://github.com/mlampros/ClusterR.
#'
#'The input matrix can be in any format supported by the `DelayedArray` /
#'`beachmat` framework, including the matrix classes defined in the `Matrix`
#'package and the `HDFMatrix` class.
#'
#'There are two possible initializations.
#'
#'\strong{kmeans++}: kmeans++ initialization.
#'
#'\strong{random}: random selection of data rows as initial centroids.
#'
#'@references Sculley, D., 2010, April. Web-scale k-means clustering. In
#'Proceedings of the 19th international conference on World wide web (pp.
#'1177-1178). ACM.
#'
#'Arthur, D. and Vassilvitskii, S., 2007, January. k-means++: The advantages of
#'careful seeding. In Proceedings of the eighteenth annual ACM-SIAM symposium
#'on Discrete algorithms (pp. 1027-1035). Society for Industrial and Applied
#'Mathematics.
#'
#'@examples
#'data = matrix(1:30,nrow = 10)
#'mini_batch(data, 2, 10, 10)
#'
#' @export
mini_batch <- function(data, clusters, batch_size, max_iters, num_init = 1L, init_fraction = 1.0, initializer = "kmeans++", compute_labels = TRUE, calc_wcss = FALSE, early_stop_iter = 10L, verbose = FALSE, CENTROIDS = NULL, tol = 1e-4) {
    .Call(`_mbkmeans_mini_batch`, data, clusters, batch_size, max_iters, num_init, init_fraction, initializer, compute_labels, calc_wcss, early_stop_iter, verbose, CENTROIDS, tol)
}

context("Test mbkmeans function.")
set.seed(13124)

test_that("mini_batch gives the same results on the iris data", {
    data(iris)
    irismat <- as.matrix(iris[,1:4])

    irisres <- list(centroids = matrix(data = c(6.95, 4.983333, 5.8, 3.45,
                                                3.383333, 2.65, 5.9, 1.433333,
                                                4.25, 2.3, 0.2, 1.25),
                                       nrow = 3, ncol=4),
                    WCSS_per_cluster = c(30.235, 15.42333, 46.3725),
                    best_initialization = 1,
                    iters_per_initialization = matrix(10),
                    Clusters =   c(2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
                                   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
                                   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 1, 3, 3, 3, 3,
                                   3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
                                   3, 1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
                                   3, 3, 3, 3, 3, 1, 3, 1, 1, 1, 1, 3, 1, 1, 1, 1, 1, 1, 3,
                                   3, 1, 1, 1, 1, 3, 1, 3, 1, 3, 1, 1, 3, 3, 1, 1, 1, 1, 1,
                                   3, 3, 1, 1, 1, 3, 1, 1, 1, 3, 1, 1, 1, 3, 1, 1, 3))

    set.seed(1)
    mb <- mini_batch(irismat, clusters=3, batch_size = 10,
                     max_iters = 10, init_fraction = 0.25)

    expect_true(all(mb$centroids - irisres$centroids <1e4))
    expect_true(all(mb$Clusters - irisres$Clusters <1e4))
    expect_equal(mb$best_initialization, irisres$best_initialization)
    expect_equal(mb$iters_per_initialization, irisres$iters_per_initialization)

    # use mbkmeans
    set.seed(1)
    mbk <- mbkmeans(t(irismat), clusters=3, batch_size = 10,
                    max_iters = 10, init_fraction = 0.25)

    expect_true(all(mbk$centroids - irisres$centroids <1e4))
    expect_true(all(mbk$Clusters - irisres$Clusters <1e4))
    expect_equal(mbk$best_initialization, irisres$best_initialization)
    expect_equal(mbk$iters_per_initialization, irisres$iters_per_initialization)

})

test_that("WCSS calculation is correct", {
    data(iris)
    irismat <- as.matrix(iris[,1:4])

    mb <- mini_batch(irismat, clusters=3, batch_size = 10,
                     max_iters = 100, init_fraction = 0.25, calc_wcss = TRUE)
    expect_equal(compute_wcss(mb$Clusters, mb$centroids, irismat), mb$WCSS_per_cluster)

    km <- kmeans(irismat, 3)
    expect_equal(compute_wcss(km$cluster, km$centers, irismat), km$withinss)
})

test_that("clustering is accurate", {
    data(iris)
    irismat <- as.matrix(iris[,1:4])

    ## when starting from kmeans solution it should not move

    km <- kmeans(irismat, 3)

    mb <- mini_batch(irismat, clusters=3, batch_size = nrow(irismat),
                     max_iters = 100, init_fraction = 0.25,
                     CENTROIDS = km$centers)
    expect_equal(mb$Clusters, km$cluster)

})

test_that("all mini_batch methods give same result", {
  library(Matrix)
  library(HDF5Array)

  set.seed(123)
  m0 <- matrix(rnorm(100), ncol=10)
  m1 <- as(m0, "sparseMatrix")
  M2 <- as(m0, "HDF5Matrix")          # HDF5Matrix instance
  M3 <- M2 + 1 - 1  # DelayedMatrix instance

  set.seed(123)
  res0 <- mini_batch(m0, clusters=3, batch_size=10, max_iters = 10)

  set.seed(123)
  res1 <- mini_batch(m1, clusters=3, batch_size=10, max_iters = 10)

  set.seed(123)
  res2 <- mini_batch(M2, clusters=3, batch_size=10, max_iters = 10)

  set.seed(123)
  res3 <- mini_batch(M3, clusters=3, batch_size=10, max_iters = 10)

  expect_equal(res0, res1)
  expect_equal(res0, res2)
  expect_equal(res0, res3)

})

context("Test mbkmeans function.")
set.seed(13124)

test_that("all mbkmeans methods give same result", {
    library(SummarizedExperiment)
    library(SingleCellExperiment)
    se <- SummarizedExperiment(matrix(rpois(600, lambda=5), nrow=10, ncol=60),
                               colData = data.frame(bio = gl(2, 30)))

    sce <- as(se, "SingleCellExperiment")

    set.seed(1)
    m_se <- mbkmeans(se, clusters=3)
    set.seed(1)
    m_sce <-  mbkmeans(sce, reduceMethod = NA, clusters=3)
    set.seed(1)
    m_m <- mbkmeans(assay(sce), clusters=3)
    set.seed(1)
    mb <- mini_batch(t(assay(sce)), clusters=3, batch_size = ncol(sce),
                     max_iters = 100, init_fraction = 1)

    expect_equal(m_se, m_sce)
    expect_equal(m_se, m_m)
    expect_equal(m_se, mb)
    
    library(Matrix)
    library(HDF5Array)
    
    set.seed(123)
    m0 <- matrix(rnorm(100), ncol=10)
    m1 <- as(m0, "sparseMatrix")
    M2 <- as(m0, "HDF5Matrix")          # HDF5Matrix instance
    M3 <- M2 + 1 - 1 # DelayedMatrix instance
    
    set.seed(123)
    res0 <- mbkmeans(m0, clusters=3)

    set.seed(123)
    res1 <- mbkmeans(m1, clusters=3)
    
    set.seed(123)
    res2 <- mbkmeans(M2, clusters=3)

    set.seed(123)
    res3 <- mbkmeans(M3, clusters=3)
    
    expect_equal(res0, res1)
    expect_equal(res0, res2)
    expect_equal(res0, res3)
    
})

test_that("Length of cluster labels is the same as number of obs", {
    library(HDF5Array)
    set.seed(123)
    m0 <- matrix(rnorm(2.3e7), ncol=1000)
    M2 <- as(m0, "HDF5Matrix")
    
    k <- 15
    batch <- 500
    
    res_1kcell <- mbkmeans(x=M2,
                           clusters=k, 
                           batch_size = batch, 
                           num_init=1, 
                           max_iters=100,
                           verbose=FALSE,
                           compute_labels=TRUE)
    
    expect_equal(length(res_1kcell$Clusters), NCOL(M2))
    
    fit <- mini_batch(data = t(M2), clusters = k,
                      batch_size = batch, max_iters = 100,
                      num_init = 1,
                      init_fraction = batch/NCOL(M2),
                      initializer = "kmeans++",
                      compute_labels = FALSE,
                      calc_wcss = FALSE,
                      early_stop_iter = 10,
                      verbose = FALSE,
                      CENTROIDS = NULL, tol = 1e-4)
    
    # the following creates 2k labels instead of 1k
    pred_r_1k <- predict_mini_batch_r(data = M2,
                                      centroids = fit$centroids)
    
    expect_equal(length(pred_r_1k), NCOL(M2))
    
    # the following works fine
    pred_c_1k <- predict_mini_batch(data = t(M2),
                                      CENTROIDS = fit$centroids)
    
    expect_equal(length(pred_c_1k), NCOL(M2))
    
    expect_equal(pred_r_1k, pred_c_1k)
    
})

#' @rdname mbkmeans
setGeneric(
    name = "mbkmeans",
    def = function(x, ...) standardGeneric("mbkmeans")
)

#' Cluster rows of a matrix
#'
#' Cluster rows of a matrix-like object with a variety of algorithms.
#'
#' This function is deprecated. Please use the \code{clusterRows} function in
#' the \code{bluster} Bioconductor package.
#' @name clusterRows
NULL

#' blocksize
#'
#' Return the maximum number of rows to use based on the amount of ram memory.
#'
#' @param data matrix-like object.
#' @param ram the max amount of ram (in bytes) to use.
#'
#' @importFrom benchmarkme get_ram
#'
#' @return  Numeric value of the maximum number of rows.
#'
#' @examples
#' data <- matrix(NA, nrow = 100, ncol=1000)
#' blocksize(data, ram=1e6)
#' @export
blocksize<-function(data, ram = get_ram()){
    if(is.na(ram)) {
        warning("Unable to determine available RAM. Setting it to 1GB.")
        ram <- 1e9
    }
    result<-min(floor(as.numeric(ram)/(2*8*nrow(data))), ncol(data))
    result
}

context("Test blocksize function.")
set.seed(13124)

test_that("blocksize works with and without ram argument", {
    data(iris)
    irismat <- as.matrix(iris[,1:4])

    expect_silent(b1 <- blocksize(irismat, ram = 1e9))
    expect_warning(b2 <- blocksize(irismat, ram = NA))

    expect_equal(b1, b2)
})

#' Compute labels for mini-batch k-means
#'
#' Given a data matrix and a centroid matrix, it assigns each data point to the
#' closest centroid, using block processing.
#'
#' @param data a matrix-like object with features in row and samples in columns.
#' @param centroids a matrix with the coordinates of the centroids.
#' @param BPPARAM for parallel computations. See the `BiocParallel` package.
#' @param ... passed to `blockApply`.
#'
#' @return a vector of cluster labels for each observation.
#'
#' @importFrom DelayedArray blockApply colSums colAutoGrid
#' @importFrom BiocParallel SerialParam
#' @export
#'
#' @examples
#'
#' data(iris)
#' km <- mini_batch(as.matrix(iris[,1:4]), clusters = 3,
#'                  batch_size = 10, max_iters = 100)
#' predict_mini_batch_r(t(as.matrix(iris[,1:4])), km$centroids)
predict_mini_batch_r <- function(data, centroids,
                                 BPPARAM=BiocParallel::SerialParam(),
                                 ...) {
    unlist(blockApply(data, all_labels, centroids = centroids,
                      grid = colAutoGrid(data),
                      BPPARAM = BPPARAM, ...))
}

one_centroid <- function(x, data) {
    colSums((x - data)^2)
}

all_labels <- function(data, centroids) {
    ss <- apply(centroids, 1, one_centroid, data = data)
    apply(ss, 1, which.min)
}

context("Test chunk dims.")

test_that("mini_batch gives the same results on different chunks", {
    library(SummarizedExperiment)
    library(HDF5Array)
    data(iris)
    irismat <- as.matrix(iris[,1:4])

    sce <- SummarizedExperiment(assays = list(iris = t(irismat)))
    saveHDF5SummarizedExperiment(sce, dir = "tmp1", replace = TRUE)
    saveHDF5SummarizedExperiment(sce, dir = "tmp2", chunkdim = c(nrow(sce), 1),
                                 replace = TRUE)
    saveHDF5SummarizedExperiment(sce, dir = "tmp3", chunkdim = c(1, nrow(sce)),
                                 replace = TRUE)

    tmp1 <- loadHDF5SummarizedExperiment("tmp1")
    tmp2 <- loadHDF5SummarizedExperiment("tmp2")
    tmp3 <- loadHDF5SummarizedExperiment("tmp3")

    set.seed(13124)
    system.time(mbk1 <- mbkmeans(tmp1, clusters = 3))
    set.seed(13124)
    system.time(mbk2 <- mbkmeans(tmp2, clusters = 3))
    set.seed(13124)
    system.time(mbk3 <- mbkmeans(tmp3, clusters = 3))

    expect_equal(mbk1, mbk2)
    expect_equal(mbk1, mbk3)
})

---
title: "An introduction to mbkmeans"
author: "Yuwei Ni, Davide Risso, Stephanie Hicks, and Elizabeth Purdom"
date: "Last modified: November 7, 2020; Compiled: `r format(Sys.time(), '%B %d, %Y')`"
bibliography: biblio.bib
output:
  BiocStyle::html_document:
    toc: true
vignette: >
  %\VignetteEncoding{UTF-8}
---

<!--
%\VignetteEngine{knitr::rmarkdown}
%\VignetteIndexEntry{mbkmeans vignette}
-->

# Installation

To install the package, please use the following.

```{r, eval=FALSE}
if (!requireNamespace("BiocManager"))
    install.packages("BiocManager")
BiocManager::install("mbkmeans")
```

# Introduction

This vignette provides an introductory example on how
to work with the `mbkmeans` package, which contains an 
implementation of the mini-batch k-means algorithm 
proposed in [@sculley2010web] for large single-cell
sequencing data. This algorithm runs the  _k_means iterations on small, random subsamples of data (``mini batches'') rather than the entire dataset. This both speeds up computation and reduces the amount of data that must be contained in memory.

The main function to be used by the users is `mbkmeans`.
This is implemented as an S4 generic and methods are
implemented for `matrix`, `Matrix`, `HDF5Matrix`, 
`DelayedMatrix`, `SummarizedExperiment`, and 
`SingleCellExperiment`.

Most of this work was inspired by the 
`MiniBatchKmeans()` function implemented in the 
`ClusterR` R package and we re-use many of the
C++ functions implemented there.

Our main contribution here is to provide an 
interface to the `DelayedArray` and `HDF5Array`
packages so that `mbkmeans` algorithm only brings into memory at any time the "mini-batches" needed for the current iteration, and never requires loading the entire dataset into memory. This allows the user to run the mini-batch 
_k_-means algorithm on very large datasets that do not fit 
entirely in memory.  `mbkmeans` also runs seamlessly on in-memory matrices provided by the user, and doing so significantly increases the speed of the algorithm compared to the standard _k_-means algorithm ([@Hicks2020]). A complete comparison of `mbkmeans` with other implementations can be found in [@Hicks2020].


The motivation for this work is the clustering of 
large single-cell RNA-sequencing (scRNA-seq) datasets, 
and hence the main focus is on Bioconductor's 
`SingleCellExperiment`and `SummarizedExperiment` 
data container. For this reason, `mbkmeans` assumes a 
data representation typical of genomic data, in which 
genes (variables) are in the rows and cells 
(observations) are in the column. This is contrary to 
most other statistical applications, and notably to 
the `stats::kmeans()` and `ClusterR::MiniBatchKmeans()`
functions that assume observations in rows.


We provide a lower-level `mini_batch()` function that 
expects observations in rows and is expected to be a direct
replacement of `ClusterR::MiniBatchKmeans()` for on-disk 
data representations such as `HDF5` files. The rest of 
this document shows the typical use case through the 
`mbkmeans()` interface; users interested in the
`mini_batch()` function should refer to its manual page.

## Example dataset

```{r options, include=FALSE, message=FALSE, warning=FALSE}
knitr::opts_chunk$set(cache=FALSE, error=FALSE, message=FALSE, warning=FALSE)
```

To illustrate a typical use case, we use the 
`pbmc4k` dataset of the 
[`TENxPBMCData` package](https://bioconductor.org/packages/release/data/experiment/html/TENxPBMCData.html). 
This dataset contains a set of about 4,000 cells from
peripheral blood from a healthy donor and is expected 
to contain many types or clusters of cell.

We would note that `mbkmeans` is designed for very large datasets, in particular for datasets large enough that either the full data matrix can't be held in memory, or the computations necessary cannot be computed. This vignette works with a small dataset to allow the user to quickly follow along and understand the commands. [@Hicks2020] shows the example of using `mbkmeans` with a dataset with 1.3 million cells, which is a more appropriate size for observing improved memory usage and improved speed. 

First, we load the needed packages.

```{r packages, message=FALSE, warning=FALSE}
library(TENxPBMCData)
library(scater)
library(SingleCellExperiment)
library(mbkmeans)
library(DelayedMatrixStats)
```

Note that in this vignette, we do not aim at identifying 
biologically meaningful clusters (that would 
entail a more sophisticated normalization of data and 
dimensionality reduction), but instead we only aim to show
how to run mini-batch _k_-means on a HDF5-backed
matrix.

We normalize the data simply by scaling for the total 
number of counts using `scater` and select the 1,000 
most variable genes and a random set of 100 cells to 
speed-up computations.

```{r}
tenx_pbmc4k <- TENxPBMCData(dataset = "pbmc4k")

set.seed(1034)
idx <- sample(seq_len(NCOL(tenx_pbmc4k)), 100)
sce <- tenx_pbmc4k[, idx]

#normalization
sce <- logNormCounts(sce)

vars <- rowVars(logcounts(sce))
names(vars) <- rownames(sce)
vars <- sort(vars, decreasing = TRUE)

sce1000 <- sce[names(vars)[1:1000],]

sce1000
```

# `mbkmeans`

The main function, `mbkmeans()`, returns a 
list object including `centroids`, 
`WCSS_per_cluster` (where WCSS stands for 
within-cluster-sum-of-squares), 
`best_initialization`, `iters_per_initiazation`
and `Clusters`.

It takes any matrix-like object as input, such 
as `SummarizedExperiment`, `SingleCellExperiment`,
`matrix`, `DelayedMatrix` and `HDF5Matrix`. 

In this example, the input is a `SingleCellExperiment` 
object.

```{r}
res <- mbkmeans(sce1000, clusters = 5,
                reduceMethod = NA,
                whichAssay = "logcounts")
```

The number of clusters (such as _k_ in the 
_k_-means algorithm) is set through the `clusters` 
argument. In this case, we set `clusters = 5` for 
no particular reason. For `SingleCellExperiment` 
objects, the function provides the `reduceMethod` 
and `whichAssay` arguments. The `reduceMethod` argument
should specify the dimensionality reduction slot 
to use for the clustering, and the default is 
"PCA". Note that this *does not perform* PCA but 
only looks at a slot called "PCA" already stored 
in the object. Alternatively, one can specify 
`whichAssay` as the assay to use as input to 
mini-batch _k_-means. This is used only when 
`reduceMethod` option is `NA`. See `?mbkmeans` 
for more details.

## Choice of arguments

There are additional arguements in `mbkmeans()` 
that make the function more flexible 
and suitable for more situations. 

### Batch size

The size of the mini batches is set through the 
`batch_size` argument. This argument gives the size of the samples of data brought into memory at any one time, and thus constrains the memory usage of the alogrithm. 

The `blocksize()` function can be used to set the batch size to the maximum allowed by the available memory.
It considers both the number of columns in the dataset
and the amount of RAM on the current matchine to 
calculate as large of a batch size as reasonable for 
the RAM available to the session. The calculation uses 
`get_ram()` function in `benchmarkme` package. See 
the `benchmarkme` vignette for more details.

```{r}
batchsize <- blocksize(sce1000)
batchsize
```

In this simple example, because the whole data fits in memory,
the default batch size would be a single 
batch of size `r ncol(sce1000)`. 

[@Hicks2020] provides a comparison of the effect of batch size for large datasets (i.e. where the full data cannot be stored in memory).  Their results show the algorithm to be robust to a large range of batch sizes (1,000-10,000 cells): accuracy of the results were equivalent if the batch size was above 500-1000 cells, with no noticable advantage in memory usage or speed as long as the batches were smaller than roughly 10,000 cells (corresponding to a subsample of the full data matrix whose memory foot print is not noticeable on most modern computers). For this reason, we would err on the larger range of batch size  suggested in [@Hicks2020] -- 10,000 cells. Furthermore, since the initialization uses by default the same number of cells as batch size, this gives a robust sample for determining the initial start point.

We would note that it is better to make choices on batch size based on choosing the *absolute* number of cells in a batch rather than choosing batch size based on a percentage of the overall sample size. If batch size is chosen based on the percentage of cells, very large datasets will use very large batches of cells, negating the memory and speed gains. Moreover such large batches are not needed for good performance of the algorithm. The continual resampling of the batches works like stochastic gradient descent and allows for (local) minimization of the objective function [@sculley2010web] with relatively small batch sizes, as demonstrated on single-cell data in the work of [@Hicks2020]. If anything, the size of the batch should be governed more by the complexity of the problem (e.g. the number of underlying true clusters or subtypes), rather than the number of overall cells in the dataset.

### Initialization

The performance of mini-batch _k_-means greatly 
depends on the process of initialization. We 
implemented two different initialization methods:

1. Random initialization, as in regular _k_-means; 
2. `kmeans++`, as proposed in [@arthur2007k]. The default is "kmeans++". 

The percentage of data to use for the initialization
centroids is set through the `init_fraction` argument, 
which should be larger than 0 and less than 1. The
default value is given so that the proportion matches the value of `batch_size`, converted to a proportion of the data. 

```{r}
res_random <- mbkmeans(sce1000, clusters = 5, 
                reduceMethod = NA,
                whichAssay = "logcounts",
                initializer = "random")
table(res$Clusters, res_random$Clusters)
```

Note that large values of `init_fraction` will result in that proportion of the data being held in memory and used in the initialization of the `kmeans++` algorithm. Therefore, large values of `init_fraction` will require a great deal of memory -- potentially more than that used by the actual clustering algorithm. Indeed keeping the fraction used constant for larger and larger datasets will result in commiserate numbers of cells being used in the initialization and increasing the memory usage (similar to the issues discussed above for batch size). Therefore, we recommend that users keep the default and make changes to the `batch_size` parameter as desired, ensuring consistent memory usage across both the initialization stage and the actual clustering stage of the algorithm. As mentioned in the discussion above on `batch_size`, for this reason we would recommend `batch_size` on the larger range suggested as reasonable in [@Hicks2020] so as to improve the initialization.

## Running `mbkmeans` with multiple values of $k$

The main parameter to set in $k$-means and its variants is the number of clusters $k$. `mbkmeans` is quick enough to rerun the clustering algorithm with different number of clusters even in very large datasets.

Here, we apply `mbkmeans` with $k$ from 5 to 15 and select the number of clusters using the elbow method (i.e., the value that corresponds to the point of inflection on the curve). We note that this is just a rule of thumb for selecting the number of clusters, and many different methods exist to decide the appropriate value of $k$.

To speed up the computations, we will cluster on the top 20 PCs.

```{r}
sce1000 <- runPCA(sce1000, ncomponents=20)

ks <- seq(5, 15)
res <- lapply(ks, function(k) {
    mbkmeans(sce1000, clusters = k,
             reduceMethod = "PCA",
             calc_wcss = TRUE, num_init=10)
})

wcss <- sapply(res, function(x) sum(x$WCSS_per_cluster))
plot(ks, wcss, type = "b")
```

From the plot, it seems that `k = 12` could be a reasonable value.

# Comparison with _k_-means

Note that if we set `init_fraction = 1`, 
`initializer = "random"`, and `batch_size = ncol(x)`, 
we recover the classic _k_-means algorithm.

```{r}
res_full <- mbkmeans(sce1000, clusters = 5,
                     reduceMethod = NA,
                     whichAssay = "logcounts",
                     initializer = "random",
                     batch_size = ncol(sce1000))
res_classic <- kmeans(t(logcounts(sce1000)), 
                      centers = 5)
table(res_full$Clusters, res_classic$cluster)
```

Note however, that since the two algorithms start from different random initializations, they will not always converge to the same solution. Furthremore, if the algorithm uses batches smaller than the full dataset (`batch_size<ncol(x)`), then `mbkmeans` will not converge to the same result as `kmeans` even with the same initialization. This is due to the fact that `kmeans` and `mbkmeans` have different search paths for an optimum and the problem is non-convex. 

For a comparison of memory usage, speed, and accuracy of `mbkmeans` with `kmeans` for the non-trivial case where the algorithm uses batches, see [@Hicks2020], where it is demonstrated that `mbkmeans` provides significant speed improvements and uses less memory than `kmeans`, without loss of accuracy. 

# Use with bluster

The `r Biocpkg("bluster")` package provides a flexible and extensible framework for clustering in Bioconductor packages/workflows. 
The `clusterRows()` function controls dispatch to different clustering algorithms. 
In this case, we will use the mini-batch _k_-means algorithm to cluster cells into cell populations based on their principal component analysis (PCA) coordinates.

```{r}
mat <- reducedDim(sce1000, "PCA")
dim(mat)
```

In the following three scenarios, we pass the number of cluster centers as a number using the `centers=5` argument. 
In addition, we can also pass other `mbkmeans` arguments (e.g. `batch_size=10`). 
In the first two cases, we only return the cluster labels. 
However, in the third scenario, we ask for the full `mbkmeans` output.  

```{r}
library(bluster)
clusterRows(mat, MbkmeansParam(centers=5))
```

```{r}
clusterRows(mat, MbkmeansParam(centers=5, batch_size=10))
```

```{r}
clusterRows(mat, MbkmeansParam(centers=5, batch_size=10), full = TRUE)
```

While many of the times clustering will be performed on in-memory data (for example after PCA), the `clusterRows` function also accepts any matrix-like object (in-memory or on-disk data representation) that `mbkmeans` accepts. 

For example, we can cluster using `mbkmeans` on the `logcounts` assay in the `sce1000` object (similar to above), which is a `DelayedMatrix` object from the `r Biocpkg("DelayedArray")` package.

**Note**: We transpose the matrix as the `clusterRows` function expects observations along the rows and variables along the columns. 

```{r}
logcounts(sce1000)
clusterRows(t(logcounts(sce1000)), MbkmeansParam(centers=4))
```

# Working with on-disk data in Bioconductor

This vignette is focused on the use of the `mbkmeans` package.
For the users interested in learning more about working with `HDF5` files and on-disk data in Bioconductor, we recommend the excellent [DelayedArray workshop](https://petehaitch.github.io/BioC2020_DelayedArray_workshop/articles/Effectively_using_the_DelayedArray_framework_for_users.html), presented by Peter Hickey at [Bioc2020](https://bioc2020.bioconductor.org/).

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MbkmeansParam.R
\docType{class}
\name{MbkmeansParam}
\alias{MbkmeansParam}
\alias{show,MbkmeansParam-method}
\title{Mini-batch k-means clustering}
\usage{
MbkmeansParam(centers, ...)
}
\arguments{
\item{centers}{An integer scalar specifying the number of centers.
Alternatively, a function that takes the number of observations and returns the number of centers. 
Note, the \code{\link{mbkmeans}} function uses the argument \code{clusters} argument to represent this argument. 
However, we use \code{centers} to match}

\item{...}{Further arguments to pass to \code{\link{mbkmeans}}.}
}
\description{
Run the mini-batch k-means \code{\link{mbkmeans}} function with the specified
number of centers within \code{clusterRows} from the
\code{bluster} Bioconductor package.
}
\details{
This function is deprecated. Please use the \code{MbkmeansParam} function in
the \code{bluster} Bioconductor package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{predict_mini_batch}
\alias{predict_mini_batch}
\title{Predict_mini_batch}
\usage{
predict_mini_batch(data, CENTROIDS)
}
\arguments{
\item{data}{matrix-like objectcontaining numeric or
integer data (obseravtions in rows, variables in columns).}

\item{CENTROIDS}{a matrix of initial cluster centroids. The rows of the
CENTROIDS matrix should be equal to the number of clusters and the columns
should equal the columns of the data.}
}
\value{
it returns a vector with the clusters.
}
\description{
Prediction function for mini-batch k-means applied to matrix-like objects.
}
\details{
This function takes the data and the output centroids and returns the
clusters.

This implementation relies very heavily on the
\code{\link[ClusterR]{MiniBatchKmeans}} implementation. We provide the
ability to work with other matrix-like objects other than base matrices (e.g,
DelayedMatrix and HDF5Matrix) through the \code{beachmat}
library.
}
\examples{
data(iris)
km = mini_batch(as.matrix(iris[,1:4]), clusters = 3,
               batch_size = 10, max_iters = 10)
clusters = predict_mini_batch(as.matrix(iris[,1:4]),
                             CENTROIDS = km$centroids)
}
\author{
Yuwei Ni
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/kmeans.R
\name{mbkmeans}
\alias{mbkmeans}
\alias{mbkmeans,SummarizedExperiment-method}
\alias{mbkmeans,SingleCellExperiment-method}
\alias{mbkmeans,LinearEmbeddingMatrix-method}
\alias{mbkmeans,ANY-method}
\title{Mini-Batch k-means for large single cell sequencing data}
\usage{
mbkmeans(x, ...)

\S4method{mbkmeans}{SummarizedExperiment}(x, whichAssay = 1, ...)

\S4method{mbkmeans}{SingleCellExperiment}(x, reduceMethod = "PCA", whichAssay = 1, ...)

\S4method{mbkmeans}{LinearEmbeddingMatrix}(x, ...)

\S4method{mbkmeans}{ANY}(
  x,
  clusters,
  batch_size = min(500, NCOL(x)),
  max_iters = 100,
  num_init = 1,
  init_fraction = batch_size/NCOL(x),
  initializer = "kmeans++",
  compute_labels = TRUE,
  calc_wcss = FALSE,
  early_stop_iter = 10,
  verbose = FALSE,
  CENTROIDS = NULL,
  tol = 1e-04,
  BPPARAM = BiocParallel::SerialParam(),
  ...
)
}
\arguments{
\item{x}{The object on which to run mini-batch k-means. It can be a
matrix-like object (e.g., matrix, Matrix, DelayedMatrix, HDF5Matrix) with
genes in the rows and samples in the columns. Specialized methods are
defined for SummarizedExperiment and SingleCellExperiment.}

\item{...}{passed to `blockApply`.}

\item{whichAssay}{The assay to use as input to mini-batch k-means. If x is a
SingleCellExperiment, this is ignored unless \code{reduceMethod = NA}.}

\item{reduceMethod}{Name of dimensionality reduction results to use as input
to mini-batch k-means. Set to NA to use the full matrix.}

\item{clusters}{the number of clusters}

\item{batch_size}{the size of the mini batches. By default, it equals the
minimum between the number of observations and 500.}

\item{max_iters}{the maximum number of clustering iterations}

\item{num_init}{number of times the algorithm will be run with different
centroid seeds}

\item{init_fraction}{proportion of data to use for the initialization
centroids (applies if initializer is \emph{kmeans++} ). Should be a float
number between 0.0 and 1.0. By default, it uses the relative batch size.}

\item{initializer}{the method of initialization. One of \emph{kmeans++} and
\emph{random}. See details for more information}

\item{compute_labels}{logcical indicating whether to compute the final cluster
labels.}

\item{calc_wcss}{logical indicating whether the per-cluster WCSS
is computed. Ignored if `compute_labels = FALSE`.}

\item{early_stop_iter}{continue that many iterations after calculation of the
best within-cluster-sum-of-squared-error}

\item{verbose}{either TRUE or FALSE, indicating whether progress is printed
during clustering}

\item{CENTROIDS}{a matrix of initial cluster centroids. The rows of the
CENTROIDS matrix should be equal to the number of clusters and the columns
should be equal to the columns of the data}

\item{tol}{a float number. If, in case of an iteration (iteration > 1 and
iteration < max_iters) 'tol' is greater than the squared norm of the
centroids, then kmeans has converged}

\item{BPPARAM}{See the `BiocParallel` package. Only the label assignment is
done in parallel.}
}
\value{
A list with the following attributes: centroids, WCSS_per_cluster,
  best_initialization, iters_per_initialization.

a list with the following attributes: centroids, WCSS_per_cluster,
 best_initialization, iters_per_initialization
}
\description{
This is an implementation of the mini-batch k-means algorithm of
  Sculley (2010) for large single cell sequencing data with the
  dimensionality reduction results as input in the reducedDim() slot.
}
\details{
The implementation is largely based on the
  \code{\link[ClusterR]{MiniBatchKmeans}} function of the \code{ClusterR}
  package. The contribution of this package is to provide support for on-disk
  data representations such as HDF5, through the use of \code{DelayedMatrix}
  and \code{HDF5Matrix} objects, as well as for sparse data representation
  through the classes of the \code{Matrix} package. We also provide
  high-level methods for objects of class \code{SummarizedExperiment},
  \code{SingleCellExperiment}, and \code{LinearEmbeddingMatrix}.

This function performs k-means clustering using mini batches.

\strong{kmeans++}: kmeans++ initialization. Reference :
http://theory.stanford.edu/~sergei/papers/kMeansPP-soda.pdf AND
http://stackoverflow.com/questions/5466323/how-exactly-does-k-means-work

\strong{random}: random selection of data rows as initial centroids
}
\examples{
library(SummarizedExperiment)
se <- SummarizedExperiment(matrix(rnorm(100), ncol=10))
mbkmeans(se, clusters = 2)
library(SingleCellExperiment)
sce <- SingleCellExperiment(matrix(rnorm(100), ncol=10))
mbkmeans(sce, clusters = 2, reduceMethod = NA)
x<-matrix(rnorm(100), ncol=10)
mbkmeans(x,clusters = 3)

}
\references{
Sculley. Web-Scale K-Means Clustering. WWW 2010, April 26–30,
  2010, Raleigh, North Carolina, USA. ACM 978-1-60558-799-8/10/04.

https://github.com/mlampros/ClusterR
}
\author{
Lampros Mouselimis and Yuwei Ni
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blocksize.R
\name{blocksize}
\alias{blocksize}
\title{blocksize}
\usage{
blocksize(data, ram = get_ram())
}
\arguments{
\item{data}{matrix-like object.}

\item{ram}{the max amount of ram (in bytes) to use.}
}
\value{
Numeric value of the maximum number of rows.
}
\description{
Return the maximum number of rows to use based on the amount of ram memory.
}
\examples{
data <- matrix(NA, nrow = 100, ncol=1000)
blocksize(data, ram=1e6)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute_labels.R
\name{predict_mini_batch_r}
\alias{predict_mini_batch_r}
\title{Compute labels for mini-batch k-means}
\usage{
predict_mini_batch_r(
  data,
  centroids,
  BPPARAM = BiocParallel::SerialParam(),
  ...
)
}
\arguments{
\item{data}{a matrix-like object with features in row and samples in columns.}

\item{centroids}{a matrix with the coordinates of the centroids.}

\item{BPPARAM}{for parallel computations. See the `BiocParallel` package.}

\item{...}{passed to `blockApply`.}
}
\value{
a vector of cluster labels for each observation.
}
\description{
Given a data matrix and a centroid matrix, it assigns each data point to the
closest centroid, using block processing.
}
\examples{

data(iris)
km <- mini_batch(as.matrix(iris[,1:4]), clusters = 3,
                 batch_size = 10, max_iters = 100)
predict_mini_batch_r(t(as.matrix(iris[,1:4])), km$centroids)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{mini_batch}
\alias{mini_batch}
\title{Mini_batch}
\usage{
mini_batch(
  data,
  clusters,
  batch_size,
  max_iters,
  num_init = 1L,
  init_fraction = 1,
  initializer = "kmeans++",
  compute_labels = TRUE,
  calc_wcss = FALSE,
  early_stop_iter = 10L,
  verbose = FALSE,
  CENTROIDS = NULL,
  tol = 1e-04
)
}
\arguments{
\item{data}{numeric or integer matrix-like object.}

\item{clusters}{the number of clusters.}

\item{batch_size}{the size of the mini batches.}

\item{max_iters}{the maximum number of clustering iterations.}

\item{num_init}{number of times the algorithm will be run with different
centroid seeds.}

\item{init_fraction}{percentage of data to use for the initialization
centroids (applies if initializer is \emph{kmeans++} ). Should be a float
number between 0.0 and 1.0.}

\item{initializer}{the method of initialization. One of \emph{kmeans++} and
\emph{random}. See details for more information.}

\item{compute_labels}{logical indicating whether to compute the final cluster
labels.}

\item{calc_wcss}{logical indicating whether the within-cluster sum of squares
should be computed and returned (ignored if `compute_labels = FALSE`).}

\item{early_stop_iter}{continue that many iterations after calculation of the
best within-cluster-sum-of-squared-error.}

\item{verbose}{logical indicating whether progress is printed on screen.}

\item{CENTROIDS}{an optional matrix of initial cluster centroids. The rows of
the CENTROIDS matrix should be equal to the number of clusters and the
columns should be equal to the columns of the data.}

\item{tol}{convergence tolerance.}
}
\value{
a list with the following attributes:

centroids: the final centroids;

WCSS_per_cluster (optional): the final per-cluster WCSS.

best_initialization: which initialization value led to the best WCSS
solution;

iters_per_initialization: number of iterations per each initialization;

Clusters (optional): the final cluster labels.
}
\description{
Mini-batch-k-means for matrix-like objects
}
\details{
This function performs k-means clustering using mini batches. It was
inspired by the implementation in https://github.com/mlampros/ClusterR.

The input matrix can be in any format supported by the `DelayedArray` /
`beachmat` framework, including the matrix classes defined in the `Matrix`
package and the `HDFMatrix` class.

There are two possible initializations.

\strong{kmeans++}: kmeans++ initialization.

\strong{random}: random selection of data rows as initial centroids.
}
\examples{
data = matrix(1:30,nrow = 10)
mini_batch(data, 2, 10, 10)

}
\references{
Sculley, D., 2010, April. Web-scale k-means clustering. In
Proceedings of the 19th international conference on World wide web (pp.
1177-1178). ACM.

Arthur, D. and Vassilvitskii, S., 2007, January. k-means++: The advantages of
careful seeding. In Proceedings of the eighteenth annual ACM-SIAM symposium
on Discrete algorithms (pp. 1027-1035). Society for Industrial and Applied
Mathematics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{compute_wcss}
\alias{compute_wcss}
\title{Compute Whithin-Cluster Sum of Squares}
\usage{
compute_wcss(clusters, cent, data)
}
\arguments{
\item{clusters}{numeric vector with the cluster assignments.}

\item{cent}{numeric matrix with the centroids (clusters in rows, variables
in columns).}

\item{data}{matrix-like object containing the data (numeric or integer).}
}
\value{
A numeric vector with the value of WCSS per cluster.
}
\description{
Given a vector of cluster labels, a matrix of centroids, and a dataset, it
computes the WCSS.
}
\examples{
data = matrix(1:30,nrow = 10)
cl <- mini_batch(data, 2, 10, 10)
compute_wcss(cl$Clusters, cl$centroids, data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R
\name{clusterRows}
\alias{clusterRows}
\title{Cluster rows of a matrix}
\description{
Cluster rows of a matrix-like object with a variety of algorithms.
}
\details{
This function is deprecated. Please use the \code{clusterRows} function in
the \code{bluster} Bioconductor package.
}
